/**********
 * Copyright (c) 2003-2004 Greg Parker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GREG PARKER ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **********/

#include "includes.h"
#include "vt100.h"
#include "data/prefs.h"
#include "rsrc/rsrc.h"

#include "putty.h"

display *gdd = NULL;

static int inited = 0;

static void vt100_set_hi_bounds(display *disp, RectangleType bounds) VT100_SEGMENT;
static Boolean read_font_from_prefs(display *disp) VT100_SEGMENT;
static void vt100_read_prefs(display *disp) VT100_SEGMENT;


static Boolean read_font_from_prefs(display *disp)
{
    uint32_t value = PrefsGetInt(prefTerminalFont, fontDefault);

    if (value == font4x6  &&  gdd->font != NanoFontID) {
        gdd->font = NanoFontID;
        gdd->font_width = 4;
        gdd->font_height = 6;
        return true;
    } 
    else if (value == font6x10  &&  gdd->font != MediumFontID) {
        gdd->font = MediumFontID;
        gdd->font_width = 6;
        gdd->font_height = 10;
        return true;
    }
    else if (gdd->font == 0) {
        // no font set and pref's value isn't understood - force to default
        gdd->font = NanoFontID;
        gdd->font_width = 4;
        gdd->font_height = 6;
        return true;
    }
    else {
        // no change
        return false;
    }
}


// bounds is lo-res
display *vt100_new(struct ssh_session_t *ss, RectangleType bounds) 
{
    if (!inited) {
        // first-time init
        init_ucs_tables();
        inited = 1;
    }

    gdd = arena_calloc(sizeof(display));

    vt100_read_prefs(gdd);

    // set some cfg options
    cfg.no_mouse_rep = 1; // xterm mouse doesn't work
    cfg.no_remote_resize = 1; // remote window manipulation doesn't work
    cfg.no_remote_charset = 1; // other fonts don't work 
    cfg.wrap_mode = 1; // wrap to next line

    cfg.bellovl = 1;      // bell overloading - if 12 beeps in 3 seconds then 
    cfg.bellovl_n = 12;   // disable bell until silent for 1 second
    cfg.bellovl_t = 3000; // fixme disable noisy bell only? switch to flash?
    cfg.bellovl_s = 1000;

    term_init();
    vt100_set_bounds(gdd, bounds);

    return gdd;
}

void vt100_free(display *disp) 
{
    term_clrsb();
    arena_free(disp);
    gdd = NULL;
}

void vt100_write(display *disp, char *bytes, long len) 
{
    if (len > 0) {
        from_backend(0, bytes, len);
        term_out();
    }
}

void vt100_update(display *disp)
{
    term_update();
}

void vt100_scroll(display *disp, int lines) 
{
    term_scroll(0, lines);
}

void vt100_activate(display *disp) 
{
    gdd->inactive = false;
    term_invalidate();
}

void vt100_deactivate(display *disp) 
{
    gdd->inactive = true;
}

// bounds is HI-RES
static void vt100_set_hi_bounds(display *disp, RectangleType bounds)
{
    int cw, ch;

    gdd->gadgetBounds = bounds;

    cw = gdd->gadgetBounds.extent.x / gdd->font_width;
    ch = gdd->gadgetBounds.extent.y / gdd->font_height;

    gdd->textBounds.topLeft = gdd->gadgetBounds.topLeft;
    gdd->textBounds.extent.x = cw * gdd->font_width;
    gdd->textBounds.extent.y = ch * gdd->font_height;
    // fixme center text area inside gadget

    term_size(ch, cw, PrefsGetInt(prefScrollbackLines,defaultScrollbackLines));
    
    term_invalidate();
}


// bounds is LO-RES
void vt100_set_bounds(display *disp, RectangleType bounds)
{
    // scale bounds to hi-res
    WinPushDrawState();
    WinSetCoordinateSystem(kCoordinatesDouble);
    WinScaleRectangle(&bounds); // convert from standard to current
    WinPopDrawState();

    vt100_set_hi_bounds(disp, bounds);
}


void vt100_size(display *disp, int *charsWide, int *charsHigh, 
                RectangleType *bounds)
{
    if (charsWide) *charsWide = gdd->textBounds.extent.x / gdd->font_width;
    if (charsHigh) *charsHigh = gdd->textBounds.extent.y / gdd->font_height;
    if (bounds) *bounds = gdd->gadgetBounds;
}


// x and y are LO-RES
// clickCount is 1 or more for mouse down, 0 for mouse up
void vt100_click(display *disp, int x, int y, int clickCount)
{
    Coord hix, hiy;
    int charx, chary;
    Mouse_Button button;
    Mouse_Action action;

    // scale point to hi-res
    WinPushDrawState();
    WinSetCoordinateSystem(kCoordinatesDouble);
    hix = WinScaleCoord(x, false); // convert from standard to current
    hiy = WinScaleCoord(y, false); // convert from standard to current
    WinPopDrawState();

    // calculate coordinate
    charx = (hix - gdd->textBounds.topLeft.x) / gdd->font_width;
    chary = (hiy - gdd->textBounds.topLeft.y) / gdd->font_height;

    // choose mouse button and mouse action
    button = MBT_LEFT;
    switch (clickCount) {
    case -1: action = MA_DRAG; break;
    case 0:  action = MA_RELEASE; break; // mouse up
    case 1:  action = MA_CLICK;   break; // mouse down
    case 2:  action = MA_2CLK;    break; // double-click
    case 3:  action = MA_3CLK;    break; // triple-click
    default: return; // ignore 4+ clicks (fixme?)
    }
    
    // fixme shift/ctrl/alt modifiers?
    // send at least one drag before release
    term_mouse(button, action, charx, chary, 0, 0, 0);
}


void vt100_deselect(display *disp)
{
    term_deselect();
}


void vt100_copy(display *disp)
{
    term_copy();
}

void vt100_paste(display *disp) 
{
    term_do_paste();
}

int vt100_selection_exists(display *disp) 
{
    return term_selection_exists();
}

static void vt100_read_prefs(display *disp)
{
    read_font_from_prefs(disp);
    WIN_read_default_colors();
    cfg.scroll_on_key = 
        PrefsGetInt(prefScrollDownOnTyping, defaultScrollDownOnTyping);
    cfg.scroll_on_disp = 
        PrefsGetInt(prefScrollDownOnActivity, defaultScrollDownOnActivity);
    // bell settings re-read every time by WIN_beep()
}

void vt100_reread_prefs(display *disp) 
{
    uint32_t value;

    vt100_read_prefs(disp);

    // resize and set scrollback
    vt100_set_hi_bounds(disp, gdd->gadgetBounds);

    // force redraw
    term_invalidate();
}

void vt100_seen_key(display *disp)
{
    seen_key_event = 1;
}
